@@ -3,7 +3,6 @@
 
 var
   assert = require('assert'),
-  fs = require('fs'),
   path = require('path'),
   inbandStandardTests = require('./inband-standard'),
   childProcess = require('./child-process').genericChildProcess,
@@ -64,7 +63,7 @@ describe('tmp', function () {
           try {
             assertions.assertDoesNotExist(stdout);
           } catch (err) {
-            rimraf.sync(stdout);
+            //rimraf.sync(stdout);
             return done(err);
           }
           done();
@@ -77,8 +76,8 @@ describe('tmp', function () {
           if (!stderr) return done(new Error('stderr expected'));
           try {
             assertions.assertExists(stdout);
-          } catch (err) {
             rimraf.sync(stdout);
+          } catch (err) {
             return done(err);
           }
           done();
@@ -91,8 +90,8 @@ describe('tmp', function () {
           if (stderr) return done(new Error(stderr));
           try {
             assertions.assertExists(stdout);
-          } catch (err) {
             rimraf.sync(stdout);
+          } catch (err) {
             return done(err);
           }
           done();
@@ -138,8 +137,8 @@ describe('tmp', function () {
               assertions.assertExists(path.join(stdout, 'symlinkme-target'), true);
             else
               assertions.assertExists(path.join(stdout, 'symlinkme-target'));
-          } catch (err) {
             rimraf.sync(stdout);
+          } catch (err) {
             return done(err);
           }
           done();
